(* bfsearch.ml     Best-first search for Slide24
 *
 * Copyright (c) 2022 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://opensource.org/licenses/MIT
 *)

(* Precompute Manhattan distance between all pairs of cells.
 *)
let manh_dist_tab =
    Bytes.init (25 * 25)
        (fun i ->
            let r0 = i / 125 in
            let c0 = (i / 25) mod 5 in
            let r1 = (i / 5) mod 5 in
            let c1 = i mod 5 in
            Char.chr (abs (r1 - r0) + abs (c1 - c0))
        )

let manh_dist cell0 cell1 =
    Bytes.get_uint8 manh_dist_tab (cell0 * 25 + cell1)


module Conf : sig
    (* A configuration of tiles.
     *)
    val cellct : int   (* Number of cells in the puzzle = 25 *)

    (* t is a bare tile configuration: 25 values in 0 .. 24.
     * cinfo is a tile configuration plus heuristic distance and path
     *)
    type t
    type cinfo

    (* In-place reinterpretation, to treat first part of cinfo as t.
     *)
    val ci_as_t : cinfo -> t
    val ci_conf_extract : t -> t

    val init : (int -> int) -> t
    val cell : t -> int -> int
    val cell_set : t -> int -> int -> unit
    val compare : t -> t -> int

    val ci_init : (int -> int) -> cinfo
    val ci_cell : cinfo -> int -> int
    val ci_cell_set : cinfo -> int -> int -> unit

    (* Cell number of the blank.
     *)
    val ci_blank : cinfo -> int
    val ci_blank_set : cinfo -> int -> unit

    (* Heuristic distance (unweighted) to goal.
     *)
    val ci_dist : cinfo -> int
    val ci_dist_set : cinfo -> int -> unit

    (* Path back to start is a sequence of 2-bit directions stored 4 per
     * byte.
     *
     * Direction refers to move of the blank: 0 = R, 1 = U, 2 = L, 3 = D
     *)
    val ci_plen : cinfo -> int
    val ci_plen_set : cinfo -> int -> unit
    val ci_path_n : cinfo -> int -> int
    val ci_path_add : cinfo -> int -> unit
    val ci_path_set : cinfo -> cinfo -> unit
    val ci_lastdir : cinfo -> int

    (* Copy just the cells of a cinfo into new cinfo.
     *)
    val ci_cells_copy : cinfo -> cinfo

    (* Extract path as list of directions.
     *)
    val ci_extract_path : cinfo -> int list
end =
struct
    let cellct = 25
    let ilength = 92

    type [@unboxed] t = Conf of bytes
    type [@unboxed] cinfo = CInfo of bytes

    let ci_as_t (CInfo bytes) = Conf bytes

    let [@inline] ci_conf_extract (Conf bytes as conf) =
        if Bytes.length bytes = cellct then
            (* A genuine conf (type t).
             *)
            conf
        else
            (* A cinfo pretending to be a conf.
             *)
            Conf (Bytes.sub bytes 0 cellct)

    let init f =
        Conf (Bytes.init cellct (fun i -> Char.chr (f i)))

    let [@inline] cell (Conf cbytes) i =
        Bytes.get_uint8 cbytes i
    let [@inline] cell_set (Conf cbytes) i b =
        Bytes.set_uint8 cbytes i b

    let compare (Conf a) (Conf b) =
        let rec c_cmp i =
            if i >= 24 then
                0 (* Last byte is redundant *)
            else
                let ab = Char.code (Bytes.unsafe_get a i) in
                let bb = Char.code (Bytes.unsafe_get b i) in
                if ab < bb then -1
                else if ab > bb then 1
                else c_cmp (i + 1)
        in
        c_cmp 0

    let ci_init f =
        CInfo (Bytes.init ilength (fun i -> Char.chr (f i)))

    let [@inline] ci_cell (CInfo ibytes) i =
        Bytes.get_uint8 ibytes i
    let [@inline] ci_cell_set (CInfo ibytes) i b =
        Bytes.set_uint8 ibytes i b

    let [@inline] ci_blank (CInfo ibytes) =
        Bytes.get_uint8 ibytes 25
    let [@inline] ci_blank_set (CInfo ibytes) b =
        Bytes.set_uint8 ibytes 25 b

    let [@inline] ci_dist (CInfo ibytes) =
        Bytes.get_uint8 ibytes 26
    let [@inline] ci_dist_set (CInfo ibytes) d =
        Bytes.set_uint8 ibytes 26 d

    let [@inline] ci_plen (CInfo ibytes) =
        Bytes.get_uint8 ibytes 27
    let [@inline] ci_plen_set (CInfo ibytes) plen =
        Bytes.set_uint8 ibytes 27 plen

    let [@inline] ci_path_n (CInfo ibytes) n =
        (* Caller warrants that n is in range, i.e., 0 <= n < plen *)
        let bytix = n lsr 2 in
        let bitix = (n land 0x3) lsl 1 in
        ((Bytes.get_uint8 ibytes (28 + bytix)) lsr bitix) land 0x3

    let ci_path_add (CInfo ibytes as cinfo) dir =
        let plen = ci_plen cinfo in
        let bytix = plen lsr 2 in
        let bitix = (plen land 0x3) lsl 1 in
        let b = Bytes.get_uint8 ibytes (28 + bytix) in
        Bytes.set_uint8 ibytes (28 + bytix) (b lor (dir lsl bitix));
        ci_plen_set cinfo (plen + 1)

    let ci_path_set (CInfo dst) (CInfo src as cisrc) =
        let plen = ci_plen cisrc in
        let bytect = ((plen + 3) lsr 2) + 1 in
        Bytes.blit src 27 dst 27 bytect

    let [@inline] ci_lastdir cinfo =
        let plen = ci_plen cinfo in
        if plen = 0 then 4 (* Out-of-range value *)
        else ci_path_n cinfo (plen - 1)

    let ci_cells_copy (CInfo bytes) =
        let bcopy = Bytes.make ilength '\000' in
        Bytes.blit bytes 0 bcopy 0 cellct;
        CInfo bcopy

    let ci_extract_path cinfo =
        let plen = ci_plen cinfo in
        let rec path_loop accum i =
            if i >= plen then
                List.rev accum
            else
                path_loop (ci_path_n cinfo i :: accum) (i + 1)
        in
        path_loop [] 0
end


(* igoal is the inverse permutation of the goal. Each goal[k] is
 * the tile in goal cell k, while each igoal[k] is tile k's goal cell.
 *
 *              goal[c] = t  iff  igoal[t] = c
 *
 *     Conf.cell goal c = t  iff  Conf.cell igoal t = c
 *
 *     goal               igoal
 *
 *      1  2  3  4  5     24  0  1  2  3
 *      6  7  8  9 10      4  5  6  7  8
 *     11 12 13 14 15      9 10 11 12 13
 *     16 17 18 19 20     14 15 16 17 18
 *     21 22 23 24  0     19 20 21 22 23
 *)
let goal = Conf.init (fun i -> (i + 1) mod Conf.cellct)
let igoal = Conf.init (fun i -> (i + 24) mod Conf.cellct)


let transpose cinfo =
    (* Transpose the given Conf.cinfo in place.
     *)
    let swap m n =
        let t = Conf.ci_cell cinfo m in
        Conf.ci_cell_set cinfo m (Conf.ci_cell cinfo n);
        Conf.ci_cell_set cinfo n t
    in
        swap 1 5; swap 2 10; swap 3 15; swap 4 20; swap 7 11;
        swap 8 16; swap 9 21; swap 13 17; swap 14 22; swap 19 23


let row_conflicts goalrows goalcols cinfo =
    (* For row and column conflicts, see: Hansson, Mayer, Yung (1985).
     * Generating Admissible Heuristics by Criticizing Solutions to
     * Relaxed Models. Tech Note CUCS-219-85, Columbia University
     * Department of Computer Science.
     *)
    let rec i_conflicts row sofar pprev prev i =
        if i >= 5 then
            sofar
        else
            let base = row * 5 in
            let tile = Conf.ci_cell cinfo (base + i) in
            if goalrows.(tile) <> row then
                i_conflicts row sofar pprev prev (i + 1)
            else
                let bi = goalcols.(tile) in
                if bi >= prev then
                    i_conflicts row sofar prev bi (i + 1)
                else if i = 4 then
                    sofar + 1
                else if bi >= pprev then
                    let c0 = i_conflicts row (sofar + 1) pprev bi (i + 1) in
                    let c1 = i_conflicts row (sofar + 1) pprev prev (i + 1) in
                    (if c0 < c1 then c0 else c1)
                else
                    let c0 = i_conflicts row (sofar + 2) (-1) bi (i + 1) in
                    let c1 = i_conflicts row (sofar + 1) pprev prev (i + 1) in
                    (if c0 < c1 then c0 else c1)
    in
    i_conflicts 0 0 (-1) (-1) 0 +
    i_conflicts 1 0 (-1) (-1) 0 +
    i_conflicts 2 0 (-1) (-1) 0 +
    i_conflicts 3 0 (-1) (-1) 0 +
    i_conflicts 4 0 (-1) (-1) 0


let linear_conflicts goalrows goalcols cinfo =
    (* Calculate linear (row and column) conflicts of the given
     * Conf.cinfo. Column conflicts are calculated as row conflicts of
     * the transpose.
     *
     * Return the number of extra moves.
     *)
    let rconf = row_conflicts goalrows goalcols cinfo in

    transpose cinfo;
    let cconf = row_conflicts goalcols goalrows cinfo in
    transpose cinfo;

    (rconf + cconf) * 2


let corner_tiles cinfo =
    (* Corner tiles heuristic. See: Korf, Richard E. and Larry A. Taylor
     * (1996). Finding Optimal Solutions to the Twenty-Four Puzzle.
     * AAAI-96 Proceedings, 1202-1207.
     *
     * Consider the goal config:
     *
     *      1  2  3  4  5
     *      6  7  8  9 10
     *     11 12 13 14 15
     *     16 17 18 19 20
     *     21 22 23 24
     *
     * If in some config: (A) tile 2 is in its goal spot; (B) a tile
     * other than 1 is in the goal spot of 1; and (C) tile 2 isn't
     * involved in a row conflict, then tile 2 will have to move out of
     * the way to get 1 to its goal. Hence can count an extra 2 moves
     * for this extra movement of tile 2.
     *
     * Tile 2 involved in row conflict means: (D) tile 1 is on top row
     * to right of tile 2, or (E) tile 3, 4, or 5 in top left corner.
     *
     * By rotation/reflection, same goes for tiles 4, 6, 10, 16, 22.
     *
     * Return the number of extra moves.
     *)
    let ct1 cbase stride =
        (* cbase = base cell; stride = delta to corner *)
        let base_gtile = cbase + 1 in        (* Correct tile for base cell *)
        if Conf.ci_cell cinfo cbase <> base_gtile then
            (* Base tile not in goal spot.
             *)
            0
        else
            let corn_tile = Conf.ci_cell cinfo (cbase + stride) in
            let corn_gtile = base_gtile + stride in (* Corner tile of goal *)
            if corn_tile <> corn_gtile &&           (* Corner is incorrect *)
                corn_tile <> 0 &&                   (* Corner is occupied *)
                (* D cases *)
                corn_tile <> base_gtile - stride &&
                corn_tile <> base_gtile - stride * 2 &&
                corn_tile <> base_gtile - stride * 3 &&
                (* E cases *)
                Conf.ci_cell cinfo (cbase - stride) <> corn_gtile &&
                Conf.ci_cell cinfo (cbase - stride * 2) <> corn_gtile &&
                Conf.ci_cell cinfo (cbase - stride * 3) <> corn_gtile then
                2
            else
                0
    in
    ct1 1 (-1) + ct1 3 1 + ct1 5 (-5) +
        ct1 9 (-5) + ct1 15 5 + ct1 21 (-1)


let distance goalrows goalcols cinfo =
    (* Estimate how far the Conf.cinfo is from being in the goal
     * configuration. goalrows and goalcols map from tile numbers to
     * their row and column in the goal (respectively), with the blank
     * mapping to -1 as a special case.
     *
     * Sum up the Manhattan distances of tiles from their goal spots,
     * and add in linear and corner tile conflicts.
     *)
    let rec manh_loop i accum =
        if i >= Conf.cellct then
            accum
        else
            let tilei = Conf.ci_cell cinfo i in
            if tilei = 0 then
                manh_loop (i + 1) accum (* No tile there *)
            else
                let goali = Conf.cell igoal tilei in
                let accum' = accum + manh_dist i goali in
                manh_loop (i + 1) accum'
    in
    let manhattan = manh_loop 0 0 in
    let l_conflicts = linear_conflicts goalrows goalcols cinfo in
    let c_tiles = corner_tiles cinfo in
    manhattan + l_conflicts + c_tiles


let compare_cinfo a b =
    (* Compare cinfos by weighted heuristic metric. For equal metrics
     * compare lexicographically by the contained conf.
     *
     * The current weight is 2.25, giving metric (2.25 * dist + plen).
     * dist is the heuristic distance to goal; plen is the path length
     * back to start. The claim is that this gives solutions within 2.25
     * times the optimal.
     *
     * In one test of 100 random configs for which optimal solutions are
     * known, this code gives a mean solution length of 144 vs. the
     * optimal mean of 103. It seems pretty good.
     *
     * This function takes around 9% of the runtime. So we use the
     * equivalent metric of (9 * dist + 4 * plen), which turns out to be
     * a little faster to compute.
     *)
    let da = Conf.ci_dist a in
    let ma = (da lsl 3) + da + (Conf.ci_plen a lsl 2) in
    let db = Conf.ci_dist b in
    let mb = (db lsl 3) + db + (Conf.ci_plen b lsl 2) in
    if ma < mb then -1
    else if ma > mb then 1
    else if da < db then -1
    else if da > db then 1
    else Conf.(compare (ci_as_t a) (ci_as_t b))


(* Frontier: set of reached but as yet unexamined configurations ordered
 * by metric (then by conf for equal metrics). This is used as a
 * priority queue via CISetByMetric.min_elt.
 *)
module CISetByMetric =
    Set.Make(struct type t = Conf.cinfo let compare = compare_cinfo end)


(* Visited: set of already examined configurations.
 *)
module CSet = struct
    include Set.Make(struct type t = Conf.t let compare = Conf.compare end)

    let add conf cset =
        (* When adding a disguised CInfo, strip down to Conf first to
         * save space.
         *)
        add (Conf.ci_conf_extract conf) cset
end


let cinfo_of_case goalrows goalcols case = 
    (* Construct a Conf.cinfo value from the given case.
     *)
    let cinfo =
        Conf.ci_init (fun i -> if i < Conf.cellct then case.(i) else 0)
    in
    Conf.ci_blank_set cinfo (S24defs.array_findi ((=) 0) case);
    Conf.ci_dist_set cinfo (distance goalrows goalcols cinfo);
    cinfo


let add_neighbors goalrows goalcols visited prevdir cinfo frontier =
    (* Determine the conf.cinfos reachable in one move from the given
     * Conf.cinfo and add them to the frontier. Return the new frontier.
     *
     * (You can think of this as moving the blank in the four orthogonal
     * directions R, U, L, D.)
     *)
    let blankx = Conf.ci_blank cinfo in
    let blank_row = blankx / 5 in
    let blank_col = blankx mod 5 in

    (* Inverse of previous dir is excluded. Note that prevdir can be 4,
     * which means there is no previous dir (when cinfo = start).
     *)
    let xdir = (0xf1032 lsr (prevdir lsl 2)) land 0xf in

    let add1 front dir nbr nbc nbx =
        if dir = xdir || nbr < 0 || nbr >= 5 || nbc < 0 || nbc >= 5 then (
            front
        ) else (
            let cinfo' = Conf.ci_cells_copy cinfo in
            Conf.ci_cell_set cinfo' blankx (Conf.ci_cell cinfo nbx);
            Conf.ci_cell_set cinfo' nbx 0;
            if CSet.mem (Conf.ci_as_t cinfo') visited then (
                front
            ) else (
                Conf.ci_blank_set cinfo' nbx;
                let ndist = distance goalrows goalcols cinfo' in
                Conf.ci_dist_set cinfo' ndist;
                Conf.ci_path_set cinfo' cinfo;
                Conf.ci_path_add cinfo' dir;
                CISetByMetric.add cinfo' front
            )
        )
    in

    add1
        (add1
            (add1
                (add1
                    frontier
                    0 blank_row (blank_col + 1) (blankx + 1))
                1 (blank_row - 1) blank_col (blankx - 5))
            2 blank_row (blank_col - 1) (blankx - 1))
        3 (blank_row + 1) blank_col (blankx + 5)


let search start maxexam : int list option * int * int =
    (* Return a list of directions showing the moves required to get
     * from start to goal. For analysis, also return the number of
     * configs that were examined in the process and the number of
     * configs remaining in the frontier at the end.
     *
     * start is the starting config, represented as an array of 25 ints.
     * The number of the tile (0 .. 24) in the upper left square of the
     * puzzle is at start.(0), the upper right tile is at start.(4), and
     * so on.
     *
     * The search is limited by maxexam, the maximum number of
     * configurations to be examined. If the goal can't be reached
     * within this limit, return None as the first element of the result
     *)
    let goalrows =
        Array.init Conf.cellct (fun i -> (Conf.cell igoal i) / 5)
    in
    let goalcols =
        Array.init Conf.cellct (fun i -> (Conf.cell igoal i) mod 5)
    in
    goalrows.(0) <- -1;
    goalcols.(0) <- -1;

    let rec isearch count visited frontier =
        if count >= maxexam || CISetByMetric.is_empty frontier then
            (None, count, CISetByMetric.cardinal frontier)
        else
            let besti = CISetByMetric.min_elt frontier in
            let bestc = Conf.ci_as_t besti in
            if Conf.compare bestc goal = 0 then
                (Some (Conf.ci_extract_path besti), count,
                    CISetByMetric.cardinal frontier)
            else
                let visited' = CSet.add bestc visited in
                let frontier' = CISetByMetric.remove besti frontier in
                let prevdir = Conf.ci_lastdir besti in
                let frontier'' =
                    add_neighbors goalrows goalcols visited' prevdir
                                    besti frontier'
                in
                isearch (count + 1) visited' frontier''
    in
    let cinfo = cinfo_of_case goalrows goalcols start in
    let visited = CSet.empty in
    let frontier = CISetByMetric.singleton cinfo in
    isearch 0 visited frontier
