(* s24anim.ml     Animate DOM elements
 *
 * Copyright (c) 2022 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://opensource.org/licenses/MIT
 *)
open Js_of_ocaml
open S24defs

(* Convenient type for callers: element, dx, dy
 *)
type elem_anim = Dom_html.element Js.t * float * float

(* Internal type: style, initial_x, final_x, initial_y, final_y)
 *)
type i_elem_anim =
    Anim of Dom_html.cssStyleDeclaration Js.t * float * float * float * float


let mk_anim (elem, dx, dy) =
    let style = elem##.style in
    let x0 = Pselcss.px_length style##.left in
    let y0 = Pselcss.px_length style##.top in
    Anim (style, x0, x0 +. dx, y0, y0 +. dy)


let ease frame framect =
    let t = float_of_int frame /. float_of_int (framect - 1) in
    if t <= 0.5 then 2.0 *. t *. t
    else 1.0 -. 2.0 *. (1.0 -. t) *. (1.0 -. t)


let rec i_animate anims frame framect final_cb =
    if frame >= framect - 1 then (
        let fz (Anim (style, _, xn, _ , yn)) =
            style##.left := Pselcss.length_px xn;
            style##.top := Pselcss.length_px yn
        in
        List.iter fz anims;
        final_cb ()
    ) else (
        let fract = ease frame framect in
        let v v0 vn = fround (v0 +. fract *. (vn -. v0)) in
        let up (Anim (style, x0, xn, y0, yn)) =
            style##.left := Pselcss.length_px (v x0 xn);
            style##.top := Pselcss.length_px (v y0 yn)
        in
        List.iter up anims;
        let next () =
            i_animate anims (frame + 1) framect final_cb
        in
        ignore (Dom_html.window##setTimeout (Js.wrap_callback next) 50.0)
    )


let animate_elems anims framect final_cb =
    (* Animate a rigid motion of the elements from their current
     * position, each with its own dx and dy. When the animation is
     * finished, call final_cb.
     *)
    i_animate (List.map mk_anim anims) 0 framect final_cb


let animate_bloc elements dx dy framect final_cb =
    (* Animate a rigid motion of the given elements, moving each from
     * its current position to one modified by the given dx and dy. When
     * the animation is finished, call final_cb.
     *)
    let anims = List.map (fun elem -> mk_anim (elem, dx, dy)) elements in
    i_animate anims 0 framect final_cb
