(* slide24ctlr.ml     Controller for Slide24
 *
 * Copyright (c) 2022 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://opensource.org/licenses/MIT
 *)
open Js_of_ocaml
open S24defs

(* The app can be in five states
 *)
type state =
    | Normal      (* Under user control *)
    | Solving     (* Calculating solution: CPU intensive *)
    | Discouraged (* Under user control, but unable to solve *)
    | Moving      (* Animating a user-initiated tile movement *)
    | Animating   (* Animating solution of puzzle, limited user control *)


(* Short names.
 *)
type pbutton = Pselbutton.t               (* Fancy button *)
type button = Dom_html.buttonElement Js.t (* DOM button  *)
type div = Dom_html.divElement Js.t
type canvas = Dom_html.canvasElement Js.t

(* Div and Pselbutton.t placeholders.
 *)
let notdiv : div = Dom_html.(createDiv document)
let nopbut = Pselbutton.create ()


type t = {
    parent : div;
    shuffle_descs : shufd list;     (* Descriptors of shuffle buttons *)
    vwd : float;                    (* Width of display area *)
    vht : float;                    (* Height of display area *)
    mutable topmargin : float;      (* Margin above tiles *)
    mutable vsep : float;           (* Separation betw tiles and buttons *)
    mutable hsep : float;           (* Horiz separation betw buttons *)
    mutable tileside : float;       (* Side of square tiles *)
    mutable borderwd : float;       (* Width of button border *)
    mutable tile_fontsz : float;    (* Font size for tiles *)
    mutable button_fontsz1 : float; (* Font size for buttons, short text *)
    mutable button_fontsz2 : float; (* Font size for buttons, longer text *)
    mutable shuffle_bs : pbutton list; (* Shuffle buttons *)
    mutable solve_b : pbutton;      (* Solve button *)
    mutable shuffle_texts : div list; (* Shuffle button texts *)
    mutable solve_text: div;        (* Solve button text *)
    mutable tiles : button array;   (* Sliding tiles *)
    mutable config : config;        (* Current configuration of tiles *)
    mutable solution : config list;
    mutable state : state;
}

(* Configuration of tiles when the puzzle is solved.
 *)
let solved = Array.init 25 (fun n -> (n + 1) mod 25)


let make_background side (bgcolor : Pselcss.color) =
    (* Make a background canvas to slice apart into individual tile
     * backgrounds.
     *)
    let inset = 4.0 in
    let g = 27 in
    let canv = Dom_html.createCanvas Dom_html.document in
    canv##.width := int_of_float side;
    canv##.height := int_of_float side;
    let ctx2d = canv##getContext Dom_html._2d_ in
    ctx2d##.fillStyle := (bgcolor :> Js.js_string Js.t);
    ctx2d##fillRect 0.0 0.0 side side;
    let rect = (inset, inset, side -. 2.0 *. inset, side -. 2.0 *. inset) in
    let spath = S24path.bzp_scale Psipath.capital_psi rect in
    let gray = CSS.Color.(js (RGBA_percent (g, g, g, 0.4))) in
    ctx2d##.fillStyle := (gray :> Js.js_string Js.t);
    S24path.draw ctx2d spath;
    canv


let [@inline] tile_is_white i =
    i mod 2 = 0

let tile_text_bdcolor i =
    if tile_is_white i then S24look.bdred else S24look.bdwhite

let tile_text_color i =
    if tile_is_white i then S24look.red else S24look.white


let frame_of_tile t i =
    let row = float_of_int (i / 5) in
    let col = float_of_int (i mod 5) in
    let left = fround ((t.vwd -. 5.0 *. t.tileside) /. 2.0) in
    (left +. t.tileside *. col,
     t.topmargin +. t.tileside *. row,
     t.tileside,
     t.tileside)


let make_tile_label t i =
    (* The label of a tile is a div containing a number.
     *)
    let ffam = "Helvetica, sans-serif" in
    let fspec = ("", "", ffam) in
    let label = Dom_html.(createDiv document) in
    let style = label##.style in
    style##.position := Js.string "absolute";
    style##.left := Pselcss.length_px (-. t.borderwd);
    style##.top := Pselcss.length_px (-. t.borderwd);
    style##.zIndex := Js.string "2";
    style##.width := Pselcss.length_px t.tileside;
    style##.height := Pselcss.length_px t.tileside;
    if t.tile_fontsz = 0.0 then
        t.tile_fontsz <-
            S24look.font_size fspec "24" (t.tileside /. 2.0) 15.0;
    style##.font := (S24look.font_spec fspec t.tile_fontsz);
    style##.lineHeight := Pselcss.length_px t.tileside;
    style##.textAlign := Js.string "center";
    style##.color := (tile_text_color i :> Js.js_string Js.t);
    let jstr = Js.string (string_of_int (i + 1)) in
    let tnode = Dom_html.document##createTextNode jstr in
    Dom.appendChild label tnode;
    label


let make_tile_image t (whiteback: canvas) (redback: canvas) i =
    (* The background image of a tile is a canvas containing a square
     * subregion of one of the two given backgrounds.
     *)
    let row = float_of_int (i / 5) in
    let col = float_of_int (i mod 5) in
    let (sx, sy, sw, sh) =
        (t.tileside *. col, t.tileside *. row, t.tileside, t.tileside)
    in
    let image = Dom_html.(createCanvas document) in
    let iside = int_of_float t.tileside in
    image##.width := iside;
    image##.height := iside;
    let ictx2d = image##getContext Dom_html._2d_ in
    let back = if tile_is_white i then whiteback else redback in
    ictx2d##drawImage_fullFromCanvas back
        sx sy sw sh 0.0 0.0 t.tileside t.tileside;
    let style = image##.style in
    style##.position := Js.string "absolute";
    style##.left := Pselcss.length_px (-. t.borderwd);
    style##.top := Pselcss.length_px (-. t.borderwd);
    style##.zIndex := Js.string "0";
    image


let button_font_spec t width s =
    (* Return a specification string (CSS syntax) for the button font
     * that works in the given width for the given string. Right now we
     * have two sizes, for strings <= 10 chars and for longer.
     *)
    let ffam = "Helvetica, sans-serif" in
    let fspec = ("", "bold", ffam) in
    let tfract1 = 0.55 in
    let tfract2 = 0.85 in
    if t.button_fontsz1 = 0.0 then (
        t.button_fontsz1 <-
            S24look.font_size fspec "Shuffle" (tfract1 *. width) 15.0;
        t.button_fontsz2 <-
            S24look.font_size fspec "Couldn't Solve" (tfract2 *. width) 15.0;
    );
    let fsz =
        if String.length s <= 10 then t.button_fontsz1
        else t.button_fontsz2
    in
    S24look.font_spec fspec fsz


let make_button_label t width height =
    (* A button label is a div that will contain text.
     *)
    let label = Dom_html.(createDiv document) in
    let style = label##.style in
    style##.width := Pselcss.length_px width;
    style##.height := Pselcss.length_px height;
    style##.position := Js.string "absolute";
    style##.left := Pselcss.length_px (-. t.borderwd);
    style##.top := Pselcss.length_px (-. t.borderwd);
    style##.color := (S24look.white :> Js.js_string Js.t);
    style##.font := button_font_spec t width "Shuffle";
    style##.lineHeight := Pselcss.length_px height;
    style##.textAlign := Js.string "center";
    style##.backgroundColor := (S24look.red :> Js.js_string Js.t);
    label


let set_border t (b : button) is_white =
    (* Set the button to have a rounded border.
     *)
    let style = b##.style in
    let h = Pselcss.px_length style##.height in
    let w = Pselcss.px_length style##.width in
    (* XXX rewrite without boxSizing
     *)
    Js.Unsafe.set style (Js.string "boxSizing") (Js.string "border-box");
    let radius = fround ((min h w) *. 0.1167) in
    style##.borderRadius := Pselcss.length_px radius;
    style##.overflow := Js.string "hidden";
    style##.borderStyle := Js.string "solid";
    style##.borderWidth := Pselcss.length_px t.borderwd;
    let borco = if is_white then S24look.dkwhite else S24look.dkred in
    style##.borderColor := (borco :> Js.js_string Js.t)


let rowcol_of_number t k =
    (* Find the row and column of the tile with the given number. The
     * number can also be 0, which returns the row and column of the
     * blank.
     *)
    let ix = array_findi ((=) k) t.config in
    (ix / 5, ix mod 5)


let tile_of_rowcol t (row, col) : button =
    t.tiles.(t.config.(row * 5 + col) - 1)


let title_buttons t txs1 tx2 =
    (* Set the titles of the buttons at the bottom.
     *)
    let (width, height) = S24look.shuffle_size in
    let rec clear (tdiv: div) =
        match Js.Opt.to_option tdiv##.firstChild with
        | None -> ()
        | Some n -> Dom.removeChild tdiv n; clear tdiv
    in
    let append (tdiv : div) (tx : Dom.text Js.t) =
        let tstyle = tdiv##.style in
        let text = Js.to_string tx##.data in
        tstyle##.font := button_font_spec t width text;
        tstyle##.lineHeight := Pselcss.length_px height;
        Dom.appendChild tdiv tx
    in
    List.iter clear t.shuffle_texts;
    clear t.solve_text;
    List.iter2 append t.shuffle_texts txs1;
    append t.solve_text tx2


let enable_buttons t enab1 enab2 =
    (* Enable or disable the buttons at the bottom.
     *)
    List.iter
        (fun b -> Pselbutton.set_disabled b (not enab1))
        t.shuffle_bs;
    Pselbutton.set_disabled t.solve_b (not enab2)


let display (* t *) =
    (* Set the state of the buttons based on the state.
     *)
    let make_tx s = 
        Dom_html.document##createTextNode (Js.string s)
    in
    let make_txs descs =
        List.map (fun (Shufd (s, _, _)) -> make_tx s) descs
    in
    let shuffle_txs = ref [] in
    let solved_tx = make_tx "Solved" in 
    let solve_tx = make_tx "Solve" in 
    let working_tx = make_tx "Working ..." in 
    let couldnt_tx = make_tx "Couldn't Solve" in 
    let stop_tx = make_tx "Stop Solving" in 
    fun t ->
        if !shuffle_txs = [] then
            shuffle_txs := (make_txs t.shuffle_descs);
        match t.state with
        | Normal when t.config = solved ->
            title_buttons t !shuffle_txs solved_tx;
            enable_buttons t true false
        | Normal
        | Moving ->
            title_buttons t !shuffle_txs solve_tx;
            enable_buttons t true true
        | Solving ->
            title_buttons t !shuffle_txs working_tx;
            enable_buttons t false false
        | Discouraged ->
            title_buttons t !shuffle_txs couldnt_tx;
            enable_buttons t true false
        | Animating ->
            title_buttons t !shuffle_txs stop_tx;
            enable_buttons t false true


let rowcol_of_mevent t mev =
    (* Figure out which tile was clicked to produce the mouse event,
     * then return its position in the grid as Some (row, col). If the
     * row and column can't be determined (which shouldn't happen),
     * return None.
     *)
    let rec find_button elem =
        (* Look upward through the document tree to find the button
         * containing the event target.
         *)
        match Dom_html.tagged elem with
        | Dom_html.Button b -> Some b
        | _ ->
            match Js.Opt.to_option elem##.parentNode with
            | None -> None
            | Some n ->
                match Dom.nodeType n with
                | Dom.Element pelem ->
                    find_button (Dom_html.element pelem)
                | _ ->
                    None
    in
    match Dom_html.eventTarget mev with
    | exception Not_found -> None
    | elem ->
        match find_button elem with
        | None -> None
        | Some b ->
            let tilenum =
                int_of_string (Js.to_string b##.innerText)
            in
            try Some (rowcol_of_number t tilenum)
            with Not_found -> None


let rowcol_blocmove t (tr, tc) : int * int * (int * int) list =
    (* If the tile at (tr, tc) can move (possibly en bloc with others),
     * return the (row, col) movement vector and the (row, col) values
     * of the tiles that move. The first entry in the list is the one
     * closest to the empty spot. If movement is not possible, return a
     * null vector and empty list.
     *)
    let null = (0, 0, []) in
    let vbloc drow col a b =
        (drow, 0, List.map (fun r -> (r, col)) (udrange a b))
    in
    let hbloc dcol row a b =
        (0, dcol, List.map (fun c -> (row, c)) (udrange a b))
    in
    let er, ec = rowcol_of_number t 0 in
    if tr < er then
        if tc = ec then vbloc 1 tc (er - 1) tr else null
    else if tr > er then
        if tc = ec then vbloc (-1) tc (er + 1) tr else null
    else
        if tc < ec then
            hbloc 1 tr (ec - 1) tc
        else
            hbloc (-1) tr (ec + 1) tc


let config_tiles t framect final_cb =
    (* Animate the tiles from their current positions so they are in the
     * positions given by t.config. When the animation is finished, call
     * final_cb.
     *)
    let make_anim (i, elems) tilenum =
        if tilenum = 0 then
            (i + 1, elems)
        else
            let elem = (t.tiles.(tilenum - 1) :> Dom_html.element Js.t) in
            let style = elem##.style in
            let ini_x = Pselcss.px_length style##.left in
            let ini_y = Pselcss.px_length style##.top in
            let (fin_x, fin_y, _, _) = frame_of_tile t i in
            if ini_x <> fin_x || ini_y <> fin_y then
                (i + 1, (elem, fin_x -. ini_x, fin_y -. ini_y) :: elems)
            else
                (i + 1, elems)
    in
    let (_, ranims) = Array.fold_left make_anim (0, []) t.config in
    if ranims = [] then
        final_cb () (* No animations to do *)
    else
        S24anim.animate_elems (List.rev ranims) framect final_cb


let rec animate_solution t =
    (* Animate a calculated solution to the puzzle.
     *)
    match t.state with
    | Animating ->
        (match t.solution with
        | [] ->
            t.state <- Normal;
            display t
        | config' :: rest ->
            t.config <- config';
            let again =
                Js.wrap_callback (fun () -> animate_solution t)
            in
            let final_cb () =
                ignore (Dom_html.window##setTimeout again 450.0)
            in
            config_tiles t 6 final_cb;
            t.solution <- rest
        )
    | Solving | Discouraged | Moving | Normal ->
        (* Abandon the animated solution
         *)
        t.solution <- [];
        t.state <- Normal;
        display t


let touch_tile t mev =
    (* Touched one of the tiles.
     *)
    (match t.state with
    | Solving | Moving | Animating -> () (* Ignore touches when busy *)
    | Normal
    | Discouraged ->
        (match rowcol_of_mevent t mev with
        | None -> () (* Shouldn't happen *)
        | Some rowcol ->
            (match rowcol_blocmove t rowcol with
            | (_, _, []) ->
                () (* Can't move the tile *)
            | (dr, dc, rcs) ->
                let dx = float_of_int dc *. t.tileside in
                let dy = float_of_int dr *. t.tileside in
                let tiles = List.map (tile_of_rowcol t) rcs in
                let elems =
                    (tiles : button list :> Dom_html.element Js.t list)
                in
                let final_cb () =
                    t.state <- Normal;
                    display t
                in
                S24anim.animate_bloc elems dx dy 9 final_cb;
                ListLabels.iter rcs ~f: (
                    fun (r, c) ->
                        let ix = r * 5 + c in
                        let ix' = ix + dr * 5 + dc in
                        t.config.(ix') <- t.config.(ix);
                        t.config.(ix) <- 0
                );
                t.state <- Moving;
                display t
            )
        )
    );
    true (* JSOO wants boolean result, but not DOM spec AFAICT *)


let touch_shuffle t count absolute mev =
    (* Touched a shuffle button.
     *)
    (match t.state with
    | Solving | Moving | Animating -> () (* Ignore touches when busy *)
    | Normal
    | Discouraged ->
        if absolute then
            Array.blit solved 0 t.config 0 (Array.length t.config);
        S24defs.shuffle t.config count;
        let final_cb () =
            t.state <- Normal;
            display t
        in
        config_tiles t 15 final_cb;
        t.state <- Moving;
        display t
    );
    true (* JSOO wants boolean result, but not DOM spec AFAICT *)


let configs_of_dirs start dirs =
    let rec i_configs accum prev blank_ix = function
        | [] -> List.rev (prev :: accum)
        | dir :: dirs ->
            let blank_ix' =
                match dir with
                | 0 -> (* R *) blank_ix + 1
                | 1 -> (* U *) blank_ix - 5
                | 2 -> (* L *) blank_ix - 1
                | _ -> (* D *) blank_ix + 5
            in
            let next = Array.copy prev in
            next.(blank_ix) <- next.(blank_ix');
            next.(blank_ix') <- 0;
            i_configs (prev :: accum) next blank_ix' dirs
    in
    let blank_ix = array_findi ((=) 0) start in
    i_configs [] start blank_ix dirs


let do_solve t =
    let lconfig = Array.copy t.config in
    match Bfsearch.search lconfig 240000 with
    | (None, _, _) ->
        t.state <- Discouraged;
        display t
    | (Some dirs, _, _) ->
        let sol_configs = configs_of_dirs lconfig dirs in
        t.solution <- S24defs.coalesce_blocs sol_configs;
        t.state <- Animating;
        display t;
        animate_solution t


let touch_solve t mev =
    (* Touched the solve button.
     *)
    (match t.state with
    | Discouraged -> ()      (* We are discouraged, don't solve again *)
    | Solving | Moving -> () (* Ignore touches when busy *)
    | Normal ->
        t.state <- Solving;
        display t;
        (* Need time to redisplay the graphics before doing the heavy
         * CPU work.
         *)
        let cb = Js.wrap_callback (fun () -> do_solve t) in
        ignore (Dom_html.window##setTimeout cb 200.0)
    | Animating ->
        (* Stop solving.
         *)
        t.state <- Moving;
        display t
    );
    true (* JSOO wants boolean result, but not DOM spec AFAICT *)


let make_tiles t =
    (* Create the 24 sliding tiles.
     *)
    let act_alpha = 0.4 in (* Alpha for darkening of clicked tile *)
    t.tileside <- floor (t.vwd *. 0.1875);
    let bside = t.tileside *. 5.0 in
    let (shufflewd, shuffleht) = S24look.shuffle_size in

    t.topmargin <- 20.0;
    let hsep = max 0.0 (t.vwd -. 2.0 *. shufflewd) /. 3.0 in
    (* (Make equal space at right and left margin.)
     *)
    let fract = hsep -. floor hsep in
    let hsep' =
        if fract > 0.1 && fract < 0.5 then
            floor hsep +. 1.0
        else
            floor hsep
    in
    t.hsep <- hsep';
    let rows = float_of_int ((List.length t.shuffle_descs + 2) / 2) in
    let vsep =
        fround ((t.vht -. t.topmargin -.bside -. rows *. shuffleht) /.
                (rows +. 1.0))
    in
    let vsep' = max 0.0 vsep in
    t.vsep <- vsep';
    t.borderwd <- fround (t.tileside *. 0.0333);
    let redback = make_background bside S24look.red in
    let whiteback = make_background bside S24look.white in
    t.tiles <- Array.init 24 (
        fun i ->
            let b = Pselbutton.create () in
            Pselbutton.set_onclick b (Some (touch_tile t));
            let label = make_tile_label t i in
            let bimg = make_tile_image t whiteback redback i in
            Pselbutton.set_inner b
                Dom.[(label :> node Js.t); (bimg :> node Js.t)];
            let (x, y, w, h) = frame_of_tile t i in
            Pselbutton.set_topleft b x y;
            Pselbutton.set_width b w;
            Pselbutton.set_height b h;
            let belem = Pselbutton.element b in
            set_border t belem (tile_is_white i);
            let lstyle = label##.style in
            let norm_fn () =
                lstyle##.backgroundColor :=
                    (S24look.clear :> Js.js_string Js.t);
                lstyle##.color :=
                    (tile_text_color i :> Js.js_string Js.t)
            in
            let hili_fn () =
                let grayed =
                    Pselcss.color_bdcolor (0.0, 0.0, 0.0, act_alpha)
                in
                lstyle##.backgroundColor := (grayed :> Js.js_string Js.t);
                let (r, g, b, _) = tile_text_bdcolor i in
                let sa1 = 1.0 -. act_alpha in
                let lgrayed =
                    Pselcss.color_bdcolor (r *. sa1, g *. sa1, b *. sa1, 1.0)
                in
                lstyle##.color := (lgrayed :> Js.js_string Js.t)
            in
            Pselbutton.set_look b Pselbutton.Normal norm_fn
                (Pselbutton.Color S24look.clear);
            Pselbutton.set_look b Pselbutton.OverDown hili_fn
                (Pselbutton.Color S24look.clear);
            Dom.appendChild t.parent belem;
            belem
    );
    t.config <- Array.copy solved


let make_1_button t i onclick =
    let act_alpha = 0.4 in  (* Alpha for darkening of clicked button *)
    let dis_grayf = 0.75 in (* Amount of grayness for disabled button *)
    let b = Pselbutton.create () in
    Pselbutton.set_onclick b onclick;
    let bct = List.length t.shuffle_descs + 1 in
    let (bw, bh) = S24look.shuffle_size in
    let bx' =
        if i mod 2 = 0 then
            fround ((t.vwd -. t.hsep) /. 2.0 -. bw)
        else
            fround ((t.vwd -. t.hsep) /. 2.0 +. t.hsep)
    in
    let rows = float_of_int ((bct + 1) / 2) in
    let fxht =
        t.topmargin +. 5.0 *. t.tileside +. rows *. bh +.
            (rows -. 1.0) *. t.vsep
    in
    let tbsep = fround ((t.vht -. fxht) /. 2.0) in
    let row = float_of_int (i / 2) in
    let by' =
        t.topmargin +. 5.0 *. t.tileside +. tbsep +. row *. (bh +. t.vsep)
    in
    Pselbutton.set_topleft b bx' by';
    Pselbutton.set_width b bw;
    Pselbutton.set_height b bh;
    let dombu = Pselbutton.element b in
    let bstyle = dombu##.style in
    let tdiv = make_button_label t bw bh in
    Pselbutton.set_inner b [(tdiv :> Dom.node Js.t)];
    let tstyle = tdiv##.style in
    let norm_fn () =
        tstyle##.color := (S24look.white :> Js.js_string Js.t);
        tstyle##.backgroundColor := (S24look.red :> Js.js_string Js.t);
        bstyle##.borderColor := (S24look.dkred :> Js.js_string Js.t)
    in
    let hili_fn () =
        let a1m = 1.0 -. act_alpha in
        let graya1m = Pselcss.color_bdcolor (a1m, a1m, a1m, 1.0) in
        tstyle##.color := (graya1m :> Js.js_string Js.t);
        let reda1m = Pselcss.color_bdcolor (a1m, 0.0, 0.0, 1.0) in
        tstyle##.backgroundColor := (reda1m :> Js.js_string Js.t);
        bstyle##.borderColor := (S24look.dkred :> Js.js_string Js.t)
    in

    let disa_fn () =
        let lgray = Pselcss.color_bdcolor (0.7, 0.7, 0.7, 1.0) in
        tstyle##.color := (lgray :> Js.js_string Js.t);
        let marron = 
            Pselcss.color_bdcolor
                (1.0 -. 0.667 *. dis_grayf,
                 0.333 *. dis_grayf,
                 0.333 *. dis_grayf,
                 1.0)
        in
        tstyle##.backgroundColor := (marron :> Js.js_string Js.t);
        let marron_fonce =
            Pselcss.color_bdcolor
                (0.8 *. (1.0 -. 0.667 *. dis_grayf),
                 0.8 *. (0.333 *. dis_grayf),
                 0.8 *. (0.333 *. dis_grayf),
                 1.0)
        in
        bstyle##.borderColor := (marron_fonce :> Js.js_string Js.t)
    in
    let slook state fn =
        Pselbutton.set_look b state fn (Pselbutton.Color S24look.clear)
    in
    slook Pselbutton.Normal norm_fn;
    slook Pselbutton.OverDown hili_fn;
    slook Pselbutton.Disabled disa_fn;
    set_border t dombu false;
    Dom.appendChild t.parent dombu;
    if i < bct - 1 then (
        t.shuffle_bs <- b :: t.shuffle_bs;
        t.shuffle_texts <- tdiv :: t.shuffle_texts;
    ) else (
        t.solve_b <- b;
        t.solve_text <- tdiv;
    )


let make_buttons t =
    (* Create the buttons at the bottom.
     *)
    List.iteri
        (fun i (Shufd (_, count, absolute)) ->
            let onclick mev = touch_shuffle t count absolute mev in
            make_1_button t i (Some onclick))
        t.shuffle_descs;
    make_1_button t (List.length t.shuffle_descs) (Some (touch_solve t));
    t.shuffle_bs <- List.rev t.shuffle_bs; (* Left-to right, top to bottom *)
    t.shuffle_texts <- List.rev t.shuffle_texts;
    display t


let create (div : div) (shufds : shufd list) =
    (* Create a controller that displays in the given div. Provide
     * shuffle buttons as described by shufds.
     *)
    let dstyle = div##.style in
    let t = {
        parent = div;
        shuffle_descs =
            if shufds = [] then [Shufd ("Shuffle", 150, false)] else shufds;
        vwd = Pselcss.px_length dstyle##.width;
        vht = Pselcss.px_length dstyle##.height;
        topmargin = 0.0;
        hsep = 0.0;
        vsep = 0.0;
        tileside = 0.0;
        borderwd = 0.0;
        tile_fontsz = 0.0;
        button_fontsz1 = 0.0;
        button_fontsz2 = 0.0;
        shuffle_bs = [];
        solve_b = nopbut;
        shuffle_texts = [];
        solve_text = notdiv;
        tiles = [||];
        config = [||];
        solution = [];
        state = Normal;
    }
    in
    Random.self_init ();
    make_tiles t;
    make_buttons t;
    t


let launch (div : div) (shufds : shufd list) =
    (* Create a controller that displays in the given div but don't
     * bother to return it.
     *)
    ignore (create div shufds)
